/*
 *
 * Copyright (C) 2011 Tim Altmann, Christian Brandt
 *
 * This toolkit is based on Tim Altmann's 2011 bachelor thesis
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "webcamqtkit.hpp"
#include <sstream>
//#include <stdio.h>
#import <QTKit/QTKit.h>
#import <Cocoa/Cocoa.h>


// -----------------------------------------------------------------------------
// Objective C
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
@interface QTKitVideoGrabber : QTCaptureDecompressedVideoOutput {
   QTCaptureSession                  *mCaptureSession;
   QTCaptureDeviceInput              *mCaptureDeviceInput;
   QTCaptureDecompressedVideoOutput  *mCaptureDecompressedVideoOutput;
   CVImageBufferRef                  mCurrentImageBuffer;
}

+ (NSArray *)videoDevices;

- (id)init;
- (void)dealloc;

- (BOOL)startCaptureFromDevice:(QTCaptureDevice *)device 
                     withWidth:(int)width 
                     andHeight:(int)height;

- (BOOL)running;
- (unsigned char*)snapshot;
- (void)stop;
/*
- (void)captureOutput:(QTCaptureOutput *)captureOutput 
  didOutputVideoFrame:(CVImageBufferRef)videoFrame 
     withSampleBuffer:(QTSampleBuffer *)sampleBuffer 
       fromConnection:(QTCaptureConnection *)connection;
*/
@end



// -----------------------------------------------------------------------------
@implementation QTKitVideoGrabber
- (id)init {
   self = [super init];
   mCaptureSession = nil;
   mCaptureDeviceInput = nil;
   mCaptureDecompressedVideoOutput = nil;
   mCurrentImageBuffer = nil;
   return self;
}



// -----------------------------------------------------------------------------
- (void)dealloc {
   [mCaptureSession release];
   [mCaptureDeviceInput release];
   [mCaptureDecompressedVideoOutput release];
   CVBufferRelease(mCurrentImageBuffer);
    
   [super dealloc];
}



// -----------------------------------------------------------------------------
+ (NSArray *)videoDevices {
   NSMutableArray *results = [NSMutableArray arrayWithCapacity:3];
   [results addObjectsFromArray:[QTCaptureDevice inputDevicesWithMediaType:QTMediaTypeVideo]];
   [results addObjectsFromArray:[QTCaptureDevice inputDevicesWithMediaType:QTMediaTypeMuxed]];
   return results;
}



// -----------------------------------------------------------------------------
-(unsigned char*)snapshot {
   CVPixelBufferRef pixels=nil;
   unsigned char* pixelsData = NULL;
   int times = 100;
   while( pixels == nil && times > 0) {
      @synchronized(self){
         pixels = mCurrentImageBuffer;
         CVBufferRetain(pixels);
         
         if( pixels != nil ) {
            CVPixelBufferLockBaseAddress(pixels, 0);
            if (CVPixelBufferGetPixelFormatType(pixels) == kCVPixelFormatType_32ARGB) {
                // remove the alpha channel (workaround for bug in QTkit)
                size_t width = CVPixelBufferGetWidth(pixels);
                size_t height = CVPixelBufferGetHeight(pixels);
                size_t Buffersize=width*height*sizeof(unsigned char)*3; // one byte per channel
                pixelsData = new unsigned char[Buffersize];
                size_t npixels = width*height;
                unsigned char* src = (unsigned char*)CVPixelBufferGetBaseAddress(pixels);
                unsigned char* dest = pixelsData;
                for (size_t n=0; n < npixels; n++) {
                    *dest = *(++src); // ignore alpha, copy Red
                    *(++dest) = *(++src); // copy Green
                    *(++dest) = *(++src); // copy Blue
                    ++dest; ++src;
                }
            } else { // kCVPixelFormatType_24RGB
                size_t rowBytes = CVPixelBufferGetBytesPerRow(pixels);
                size_t height = CVPixelBufferGetHeight(pixels);
                size_t Buffersize=rowBytes*height*sizeof(unsigned char);
                pixelsData = new unsigned char[Buffersize];
                memcpy(pixelsData, (unsigned char*)CVPixelBufferGetBaseAddress(pixels), Buffersize);
            }
            CVPixelBufferUnlockBaseAddress(pixels, 0);
            CVBufferRelease(pixels);
         }
      }
      
      if( pixels == nil ){
      	 times--;
         [[NSRunLoop currentRunLoop] runUntilDate:[NSDate dateWithTimeIntervalSinceNow: 0.1]];
      }
      
   }
    
   return pixelsData;
}



// -----------------------------------------------------------------------------
-(BOOL)running {
   return [mCaptureSession isRunning];
}



// -----------------------------------------------------------------------------
-(void)stop {
    
   while( mCaptureSession != nil ){
      
      [mCaptureSession stopRunning];
      
      if( [mCaptureSession isRunning] ){
         [[NSRunLoop currentRunLoop] runUntilDate:[NSDate dateWithTimeIntervalSinceNow: 0.1]];
      }
      else {
         [mCaptureSession release];
         [mCaptureDeviceInput release];
         [mCaptureDecompressedVideoOutput release];
           
         mCaptureSession = nil;
         mCaptureDeviceInput = nil;
         mCaptureDecompressedVideoOutput = nil;
      }      
   }
}



// -----------------------------------------------------------------------------
-(BOOL)startCaptureFromDevice:(QTCaptureDevice *)device 
                    withWidth:(int)width 
                    andHeight:(int)height {

   if( device == nil ) {
      return NO;
   }
   
   NSError *error = nil;   
   
   mCaptureSession = [[QTCaptureSession alloc] init];
   if( ![device open:&error] ){
      [mCaptureSession release];
      mCaptureSession = nil;
      return NO;
   }
    
   mCaptureDeviceInput = [[QTCaptureDeviceInput alloc] initWithDevice:device];
   if (![mCaptureSession addInput:mCaptureDeviceInput error:&error]) {
      [mCaptureSession release];
      [mCaptureDeviceInput release];
      mCaptureSession = nil;
      mCaptureDeviceInput = nil;
      return NO;
   }
    
   mCaptureDecompressedVideoOutput = [[QTCaptureDecompressedVideoOutput alloc] init];
   [mCaptureDecompressedVideoOutput setDelegate:self];
   if (![mCaptureSession addOutput:mCaptureDecompressedVideoOutput error:&error]) {
      [mCaptureSession release];
      [mCaptureDeviceInput release];
      [mCaptureDecompressedVideoOutput release];
      mCaptureSession = nil;
      mCaptureDeviceInput = nil;
      mCaptureDecompressedVideoOutput = nil;
      return NO;
   }

    
   [mCaptureDecompressedVideoOutput setPixelBufferAttributes:[
      NSDictionary dictionaryWithObjectsAndKeys:
            [NSNumber numberWithDouble:width], (id)kCVPixelBufferWidthKey,
            [NSNumber numberWithDouble:height], (id)kCVPixelBufferHeightKey,
            //[NSNumber numberWithUnsignedInt:kCVPixelFormatType_24RGB], // does not work on Lion
            [NSNumber numberWithUnsignedInt:kCVPixelFormatType_32ARGB], 
               (id)kCVPixelBufferPixelFormatTypeKey,
            nil
      ]];
    
   @synchronized(self) {
      if( mCurrentImageBuffer != nil ) {
         CVBufferRelease(mCurrentImageBuffer);
         mCurrentImageBuffer = nil;
      }
   }
    
   [mCaptureSession startRunning];
   if( mCurrentImageBuffer == nil ) {
      [[NSRunLoop currentRunLoop] runUntilDate:[NSDate dateWithTimeIntervalSinceNow: 0.1]];
   }
   
    return YES;
}  



// -----------------------------------------------------------------------------
- (void)captureOutput:(QTCaptureOutput *)captureOutput 
  didOutputVideoFrame:(CVImageBufferRef)videoFrame 
     withSampleBuffer:(QTSampleBuffer *)sampleBuffer 
       fromConnection:(QTCaptureConnection *)connection {
    
   if (videoFrame == nil ) 
   {
      return;
   }
    
   CVImageBufferRef imageBufferToRelease;
   CVBufferRetain(videoFrame);
   @synchronized(self){
      imageBufferToRelease = mCurrentImageBuffer;
      mCurrentImageBuffer = videoFrame;
      CVBufferRelease(imageBufferToRelease);
   }  
}
@end



// -----------------------------------------------------------------------------
// C++ Webcam Class
// -----------------------------------------------------------------------------

namespace WebcamToolkit {

Webcam* newWebcam() {
   return new WebcamQtkit;
}

// -----------------------------------------------------------------------------
std::string WebcamQtkit::getDevices() {
   NSArray *devices = [QTKitVideoGrabber videoDevices];
   std::string deviceList;
   size_t count = 0;
   
   if (((int)[devices count]) <= 0) {
      deviceList = "No device";
   }
   else {
      std::stringstream deviceStream;

      for( QTCaptureDevice *device in devices ) {
         if (count > 0) {
            deviceStream << "[/]";
         }
         deviceStream << (char*)[[device description] UTF8String];

         count++;
      }
      deviceList = deviceStream.str();
   }

   return deviceList;
}



// -----------------------------------------------------------------------------
void* WebcamQtkit::getDevicesByName(char* deviceName) {
   NSArray *devices = [QTKitVideoGrabber videoDevices];
   QTCaptureDevice* SelDevice;
   SelDevice=NULL;
   if (((int)[devices count]) > 0){
      for(QTCaptureDevice *device in devices) {   
         SelDevice=device;
         if (strcmp(deviceName,(char*)[[SelDevice description] UTF8String])==0) {
            break;
         }
      }
   }
   
   return SelDevice;
}



// -----------------------------------------------------------------------------
WebcamQtkit::WebcamQtkit() {
   NSApplicationLoad();
   pool = [[NSAutoreleasePool alloc] init];
   [NSApplication sharedApplication];
   snap = [[QTKitVideoGrabber alloc] init];
}



// -----------------------------------------------------------------------------
WebcamQtkit::~WebcamQtkit() {
   [(QTKitVideoGrabber*)snap stop];
   [(QTKitVideoGrabber*)snap release];
   [(NSAutoreleasePool *)pool drain];
}



// -----------------------------------------------------------------------------
int WebcamQtkit::initSession(int width, int height,char* device) {
   QTCaptureDevice* SelDevice;
   if (snap!=NULL) {
      if ([(QTKitVideoGrabber*)snap running]) {   
         closeSession();
      }
   }

   SelDevice=(QTCaptureDevice*)getDevicesByName(device);

   if (SelDevice!=NULL) {
      heightImage = height;
      widthImage = width;
      if( [(QTKitVideoGrabber*)snap startCaptureFromDevice:SelDevice 
            withWidth:width andHeight:height]){
         return 1;
      }
      
      return 0;
   }
   else {
      return 0;
   }
        
}   



// -----------------------------------------------------------------------------
unsigned char* WebcamQtkit::snapshotFromWebcam() {
   unsigned char* pixels;
   unsigned char* picture;
   if ([(QTKitVideoGrabber*)snap running]) {
      pixels = [(QTKitVideoGrabber*)snap snapshot];
   }
   return pixels;
}



// -----------------------------------------------------------------------------
void WebcamQtkit::closeSession() {
   [(QTKitVideoGrabber*)snap stop];
}

} //end WebcamToolkit::

