/*
 *
 * Copyright (C) 2011 Christian Brandt
 *
 * This toolkit is based on Tim Altmann's 2011 bachelor thesis
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include "webcamv4l.hpp"

#include <cassert>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cassert>
#include <algorithm>
#include <string>
#include <sstream>
#include <iostream>
#include <stdexcept>

#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <asm/types.h>
#include <linux/videodev2.h>
#include <dirent.h>

namespace WebcamToolkit {

struct buffer {
   void* start;
   size_t length;
};



// -----------------------------------------------------------------------------
Webcam* newWebcam() {
   return new WebcamV4l;
}



// -----------------------------------------------------------------------------
WebcamV4l::WebcamV4l() {
   dev_name = NULL;
   fd = -1;
   buffers = NULL;
   n_buffers = 0;
   widthImage = 640;
   heightImage = 480;
   isInit=0;
}



// -----------------------------------------------------------------------------
WebcamV4l::~WebcamV4l() {
   closeSession();
   delete [] dev_name;
}



// -----------------------------------------------------------------------------
void WebcamV4l::errno_exit(const char* s) {
   std::cerr << s << " error " << errno << ", " << strerror(errno) << std::endl;

   throw std::runtime_error("There was an error in WebcamV4l");
}



// -----------------------------------------------------------------------------
int WebcamV4l::xioctl(int fd, int request, void* arg) {
   int r;

   do {
      r = ioctl(fd, request, arg);
   }
   while (-1 == r && EINTR == errno);

   return r;
}



// -----------------------------------------------------------------------------
inline int WebcamV4l::RGB(int min, int max,int val) {
   return std::max(std::min(val,max),min);
}



// -----------------------------------------------------------------------------
void WebcamV4l::YUY2ToRGB( unsigned char* src, int len, unsigned char* dst ) {
   int i;
   for ( i = 0; i < len; i += 4 ) {
      float C  = 1.164 * (src[i]  - 16);
      float D  = 0.813 * (src[i+1]-128);
      float D1 = 1.596 * (src[i+1]-128);
      float E  = 2.018 * (src[i+3]-128);
      float E1 = 0.391 * (src[i+3]-128);

      *dst++ = RGB( 0, 255, ( C +     E  ));   //B
      *dst++ = RGB( 0, 255, ( C - D - E1 ));   //n_buffersG
      *dst++ = RGB( 0, 255, ( C + D1     ));   //R

      C = 1.164 * ( src[i+2] - 16 );

      *dst++ = RGB( 0, 255, ( C +     E  ));   //B
      *dst++ = RGB( 0, 255, ( C - D - E1 ));   //G
      *dst++ = RGB( 0, 255, ( C + D1     ));   //R

   }
}



// -----------------------------------------------------------------------------
unsigned char* WebcamV4l::snapshotFromWebcam() {
   unsigned char* picture = NULL;
   if (isInit!=0) {
      fd_set fds;
      struct timeval tv;
      int r;

      FD_ZERO (&fds);
      FD_SET (fd, &fds);

      tv.tv_sec = 2;
      tv.tv_usec = 5;
      r = select (fd+1 , &fds, NULL, NULL, &tv);


      while (0 == r) {
         tv.tv_sec = 1;
         tv.tv_usec = 0;
         r = select (fd+1 , &fds, NULL, NULL, &tv);

         if (-1 == r) {
            return NULL;
         }
      }

      read_frame();
      read_frame();
      picture= new unsigned char[widthImage*heightImage*3];
      YUY2ToRGB((unsigned char*) buffers[0].start,buffers[0].length,picture);
   }

   return picture;
}



// -----------------------------------------------------------------------------
int WebcamV4l::read_frame() {
   struct v4l2_buffer buf;
   CLEAR (buf);

   buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
   buf.memory = V4L2_MEMORY_MMAP;
   buf.timecode.type=V4L2_TC_TYPE_30FPS;
   buf.timecode.flags=V4L2_TC_FLAG_DROPFRAME;

   if (-1 == xioctl (fd, VIDIOC_DQBUF, &buf)) {
      switch (errno) {
      case EAGAIN:
         return 0;
      case EIO:

      default:
         std::cerr << "VIDIOC_DQBUF\n";
         return 0;
      }
   }


   assert (buf.index < n_buffers);

   if (-1 == xioctl (fd, VIDIOC_QBUF,&buf)) {
      std::cerr << "VIDIOC_DQBUF\n";
      return 0;
   }

   if (-1 == xioctl (fd, VIDIOC_DQBUF, &buf)) {
      switch (errno) {
      case EAGAIN:
         return 0;
      case EIO:

      default:
         std::cerr << "VIDIOC_DQBUF\n";
         return 0;
      }
   }

   return 1;
}



// -----------------------------------------------------------------------------
int WebcamV4l::stop_capturing(void) {
   if (isInit!=0) {
      enum v4l2_buf_type type;

      type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

      if (-1 == xioctl (fd, VIDIOC_STREAMOFF, &type)) {
         std::cerr << "VIDIOC_STREAMOFF\n";
         return 0;
      }
   }
   return 1;
}



// -----------------------------------------------------------------------------
int WebcamV4l::start_capturing(void) {
   unsigned int i;
   enum v4l2_buf_type type;
   for (i = 0; i < n_buffers; ++i) {
      struct v4l2_buffer buf;

      CLEAR (buf);

      buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      buf.memory      = V4L2_MEMORY_MMAP;
      buf.index       = i;

      if (-1 == xioctl (fd, VIDIOC_QBUF, &buf)) {
         std::cerr << "VIDIOC_QBUF\n";
         return 0;
      }
   }

   type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

   if (-1 == xioctl (fd, VIDIOC_STREAMON, &type)) {
      std::cerr << "VIDIOC_STREAMON\n";
      return 0;
   }
   return 1;

}



// -----------------------------------------------------------------------------
int WebcamV4l::uninit_device(void) {
   unsigned int i;
   for (i = 0; i < n_buffers; ++i) {
      if (-1 == munmap (buffers[i].start, buffers[i].length)) {
         std::cerr << "mmap\n";
         return 0;
      }
   }

   free (buffers);
   return 1;
}



// -----------------------------------------------------------------------------
int WebcamV4l::init_mmap(void) {
   struct v4l2_requestbuffers req;

   CLEAR (req);

   req.count               = 4;
   req.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
   req.memory              = V4L2_MEMORY_MMAP;

   if (-1 == xioctl (fd, VIDIOC_REQBUFS, &req)) {
      if (EINVAL == errno) {
         std::cerr << dev_name << " does not support memory mapping\n";
         return 0;
      } else {
         std::cerr << "VIDIOC_REQBUFS\n";
         return 0;
      }
   }

   if (req.count < 2) {
      std::cerr << "Insufficient buffer memory on " << dev_name << std::endl;
      return 0;
   }

   buffers = (buffer*) calloc (req.count, sizeof (*buffers));

   if (!buffers) {
      std::cerr << "Out of memory\n";
      return 0;
   }

   for (n_buffers = 0; n_buffers < req.count; ++n_buffers) {
      struct v4l2_buffer buf;

      CLEAR (buf);

      buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      buf.memory      = V4L2_MEMORY_MMAP;
      buf.index       = n_buffers;

      if (-1 == xioctl (fd, VIDIOC_QUERYBUF, &buf)) {
         std::cerr << "VIDIOC_QUERYBUF\n";
         return 0;
      }

      buffers[n_buffers].length = buf.length;
      buffers[n_buffers].start =
         mmap (NULL ,
               buf.length,
               PROT_READ  ,
               MAP_SHARED ,
               fd, buf.m.offset);

      if (MAP_FAILED == buffers[n_buffers].start) {
         std::cerr << "mmap\n";
         return 0;
      }
   }
   return 1;
}



// -----------------------------------------------------------------------------
int WebcamV4l::init_device(int width, int height) {
   struct v4l2_capability cap;
   struct v4l2_cropcap cropcap;
   struct v4l2_crop crop;
   struct v4l2_format fmt;
   unsigned int min;

   widthImage=width;
   heightImage=height;

   if (-1 == xioctl (fd, VIDIOC_QUERYCAP, &cap)) {
      if (EINVAL == errno) {
         std::cerr << dev_name << " is no V4L2 device\n";
         return 0;
      } else {
         std::cerr << "VIDIOC_QUERYCAP\n";
         return 0;
      }
   }

   if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
      std::cerr << dev_name << " is no video capture device\n";
      return 0;
   }

   if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
      std::cerr << dev_name << " does not support streaming i/o\n";
      return 0;
   }

   CLEAR (cropcap);

   cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

   if (0 == xioctl (fd, VIDIOC_CROPCAP, &cropcap)) {
      crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      crop.c = cropcap.defrect;

      if (-1 == xioctl (fd, VIDIOC_S_CROP, &crop)) {
         switch (errno) {
         case EINVAL:
            break;
         default:
            break;
         }
      }
   }


   CLEAR (fmt);

   fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
   fmt.fmt.pix.width       = widthImage;
   fmt.fmt.pix.height      = heightImage;
   fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
   fmt.fmt.pix.field       = V4L2_FIELD_INTERLACED;

   if (-1 == xioctl (fd, VIDIOC_S_FMT, &fmt)) {
      std::cerr << "VIDIOC_S_FMT\n";
      return 0;
   }


   min = fmt.fmt.pix.width * 2;
   if (fmt.fmt.pix.bytesperline < min)
      fmt.fmt.pix.bytesperline = min;
   min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
   if (fmt.fmt.pix.sizeimage < min)
      fmt.fmt.pix.sizeimage = min;

   if (init_mmap()==0)
      return 0;
   return 1;
}



// -----------------------------------------------------------------------------
int WebcamV4l::close_device() {
   if (fd != -1) {
      if (-1 == close(fd)) {
         std::cerr << "close\n";
         return 0;
      }

      fd = -1;
   }
   return 1;
}



// -----------------------------------------------------------------------------
int WebcamV4l::open_device() {
   struct stat st;

   if (dev_name==0)
      return 0;
   if (-1 == stat (dev_name, &st)) {
      std::cerr << "Cannot identify '" << dev_name << "': " << 
         errno << ", " << strerror(errno) << std::endl;
      return 0;
   }

   if (!S_ISCHR (st.st_mode)) {
      std::cerr << dev_name << " is no device\n";
      return 0;
   }

   fd = open (dev_name, O_RDWR | O_NONBLOCK, 0);

   if (-1 == fd) {
      std::cerr << "Cannot open '" << dev_name << "': " << errno << ", " << 
         strerror(errno) << std::endl;
      return 0;
   }

   return 1;
}



// -----------------------------------------------------------------------------
int WebcamV4l::set_device(char* device) {
   if (strcmp(device,"No device")==0) {
      dev_name=NULL;
      return 0;
   }
   else {
      dev_name=getDevicesByName(device);
      return 1;
   }
}



// -----------------------------------------------------------------------------
void WebcamV4l::closeSession() {
   if (isInit!=0) {
      stop_capturing();
      uninit_device();
      close_device();
      isInit=0;
   }
}



// -----------------------------------------------------------------------------
int WebcamV4l::initSession(int width,int height,char* device) {
   if (0==set_device(device)) {
      isInit=0;
      return 0;
   }
   if (0==open_device()) {
      isInit=0;
      return 0;
   }
   if (0==init_device(width,height)) {
      isInit=0;
      return 0;
   }
   if (0==start_capturing()) {
      isInit=0;
      return 0;
   }

   isInit=1;
   return 1;
}



// -----------------------------------------------------------------------------
std::string WebcamV4l::getDevices() {
   DIR *dp = NULL;
   struct dirent *dirp = NULL;
   char dir[] ="/dev/";

   std::string deviceList;
   size_t count = 0;

   if ((dp = opendir(dir)) != NULL) {
      std::stringstream devices;
      while ((dirp = readdir(dp)) != NULL) {
         std::string devname(dirp->d_name);
         if (devname.substr(0,5).compare("video") == 0) {
            if(count > 0)
               devices << "[/]";

            devices << devname;
            count++;
         }
      }

      deviceList = devices.str();
   }

   closedir(dp);

   if (count==0)
   {
      deviceList = "No device";
   }
   return deviceList;
}



// -----------------------------------------------------------------------------
char* WebcamV4l::getDevicesByName(char* deviceName)
{
   DIR *dp = NULL;
   struct dirent *dirp = NULL;
   char dir[] ="/dev/";
   char* deviceItem = NULL;
   
   if ((dp = opendir(dir)) != NULL) {
      while ((dirp = readdir(dp)) != NULL) {
         std::string devname(dirp->d_name);
         if (devname.substr(0,5).compare("video") == 0) {
            std::string device(dir);
            device.append(devname);
            if(devname.compare(deviceName) == 0) {
               size_t length = device.length();
               deviceItem = new char[length + 1];
               device.copy(deviceItem, length);
               deviceItem[length] = 0;
               break;
            }
         }
      }
   }

   closedir(dp);
   return deviceItem;
}


} //end WebcamToolkit::

